import { useEffect, useMemo, useRef, useState } from 'react';
export const useReactiveForm = ({ fields, schema, deps = [], separator = '_', validateOnChange = false, actionOnChange, updateTriggers = [], }) => {
    const ref = useRef(null);
    const [time, reload] = useState(0);
    let form = useRef(fields);
    let validationObject = useRef(deepCopy(fields));
    const errorsMap = useRef(new Map());
    const fieldsMap = useMemo(() => {
        return flattenObject(fields, separator);
    }, [fields]);
    const action = (e) => {
        const element = e.target;
        const name = element.getAttribute('name');
        const keys = name ? name.split(separator) : [];
        findKeyAndUpdateValue(keys, form.current, element);
        findKeyAndUpdateValue(keys, validationObject.current, element);
        actionOnChange && actionOnChange(form.current);
    };
    const sub = (e) => {
        const element = e.target;
        const type = element.getAttribute('type');
        const name = element.getAttribute('name');
        if (e.type === 'blur' && (type === 'radio' || type === 'checkbox')) {
            return;
        }
        if (e.type === 'focus' && (type === 'radio' || type === 'checkbox')) {
            const elements = ref.current && ref.current.querySelectorAll(`[name="${name}"]`);
            elements && elements.forEach((e) => e.classList.add('touched'));
            return;
        }
        if (e.type === 'focus' && !element.classList.contains('touched')) {
            element.classList.add('touched');
            return;
        }
        if (type === 'radio' || type === 'checkbox') {
            const elements = ref.current && ref.current.querySelectorAll(`[name="${name}"]`);
            elements && elements.forEach((e) => e.classList.add('dirty'));
        }
        else if (e.type === 'change') {
            element.classList.add('dirty');
        }
        const keys = name ? name.split(separator) : [];
        findKeyAndUpdateValue(keys, form.current, element);
        findKeyAndUpdateValue(keys, validationObject.current, element);
        validateOnChange && dynamicValidation(name, form.current, element);
        if (updateTriggers && name && updateTriggers.indexOf(name) >= 0) {
            update(form.current);
        }
    };
    useEffect(() => {
        let selectors = [];
        let event = 'change';
        if (ref.current) {
            selectors = ref.current.querySelectorAll('[name]');
            selectors.forEach((field) => {
                const name = field.getAttribute('name');
                if (name && fieldsMap[name]) {
                    if (actionOnChange) {
                        field.addEventListener(event, action);
                    }
                    else {
                        field.addEventListener(event, sub);
                        field.addEventListener('focus', sub);
                        field.addEventListener('blur', sub);
                    }
                }
            });
        }
        return () => {
            selectors.forEach((field) => {
                field.removeEventListener(event, sub);
                field.removeEventListener('focus', sub);
                field.removeEventListener('blur', sub);
                field.removeEventListener(event, action);
            });
        };
    }, [time, fieldsMap]);
    useEffect(() => {
        deps.length && update(fields);
    }, [...deps]);
    const findKeyAndUpdateValue = (keys, obj, element, error) => {
        if (!keys.length) {
            return;
        }
        if (keys.length === 1) {
            if (error) {
                obj[keys[0]] = {
                    value: obj[keys[0]],
                    error: error,
                };
                return obj[keys[0]];
            }
            if (element) {
                if (element.getAttribute('type') === 'checkbox') {
                    if (!Array.isArray(obj[keys[0]])) {
                        obj[keys[0]] = [];
                    }
                    const index = obj[keys[0]].indexOf(element.value);
                    obj[keys[0]] = index < 0 ? [...obj[keys[0]], element.value] : obj[keys[0]].filter((v) => v !== element.value);
                }
                else if (element.getAttribute('type') === 'radio') {
                    obj[keys[0]] = typeof obj[keys[0]] === 'number' ? +element.value : element.value;
                }
                else if (element.tagName === 'SELECT') {
                    const multiple = element.getAttribute('multiple');
                    if (multiple === '' || multiple) {
                        if (!Array.isArray(obj[keys[0]])) {
                            obj[keys[0]] = [];
                        }
                        obj[keys[0]] = Array.from(element.getElementsByTagName('option'))
                            .filter((o) => o.selected)
                            .map((o) => o.value);
                    }
                    else {
                        obj[keys[0]] = element.value;
                    }
                }
                else {
                    obj[keys[0]] = element.value;
                }
            }
            return obj[keys[0]];
        }
        for (const k in obj) {
            if (obj.hasOwnProperty(k) && k === keys[0]) {
                const tmp = [...keys];
                tmp.splice(0, 1);
                return findKeyAndUpdateValue(tmp, obj[k], element, error);
            }
        }
    };
    const update = (fields) => {
        form.current = fields;
        validationObject.current = deepCopy(fields, validationObject.current);
        reload(Date.now());
    };
    const validate = () => {
        if (!schema) {
            return true;
        }
        const elements = ref.current && ref.current.querySelectorAll('input.invalid, textarea.invalid');
        elements && elements.forEach((e) => e.classList.remove('invalid'));
        try {
            schema.validateSync(form.current, { abortEarly: false });
            update(form.current);
            return true;
        }
        catch (e) {
            e.inner.forEach((item) => {
                errorsMap.current.set(item.path, e.message);
                const keys = item.path.split(/\[|].|\./);
                findKeyAndUpdateValue(keys, validationObject.current, null, item.message);
                const selector = item.path.replace(/[[.]/g, separator).replace(/]/g, '');
                const elements = ref.current && ref.current.querySelectorAll(`[name="${selector}"]`);
                elements && elements.forEach((e) => e.classList.add('invalid'));
            });
            reload(Date.now());
            return false;
        }
    };
    const dynamicValidation = (name, values, element) => {
        let path;
        path = name ? name.replace(new RegExp(separator, 'g'), '_')
            .replace(/(_)/g, (_, sign, offset) => /\d/g.test(name[offset + 1]) ? '[' : /\d/g.test(name[offset - 1]) ? ']' : sign)
            .replace(/(])/g, (_, sign, offset) => /\w/g.test(name[offset + 1]) ? '].' : sign) : '';
        const type = element.getAttribute('type');
        let shouldUpdate;
        let valid;
        const isSelect = element.getAttribute('data-select') === 'true';
        const errors = validationObject.current;
        const keys = name ? name.split(separator) : [];
        try {
            schema.validateSyncAt(path, values);
            valid = true;
            shouldUpdate = errorsMap.current.get(path) !== '' && errorsMap.current.get(path) !== undefined;
            errorsMap.current.set(path, '');
            findKeyAndUpdateValue(keys, errors, element);
        }
        catch (e) {
            valid = false;
            shouldUpdate = errorsMap.current.get(path) !== e.message;
            errorsMap.current.set(path, findKeyAndUpdateValue(keys, errors, element, e.message).error);
        }
        validationObject.current = errors;
        if (type === 'radio' || type === 'checkbox') {
            const elements = ref.current && ref.current.querySelectorAll(`[name="${name}"]`);
            elements && elements.forEach((e) => {
                valid ? e.classList.remove('invalid') : e.classList.add('invalid');
            });
        }
        else {
            valid ? element.classList.remove('invalid') : element.classList.add('invalid');
        }
        if (isSelect) {
            reload(Date.now());
        }
        else {
            shouldUpdate && reload(Date.now());
        }
    };
    const clear = () => update(fields);
    return { values: form.current, ref, update, validate, clear, errors: validationObject.current };
};
function deepCopy(obj, preserve) {
    let clone = {};
    for (const key in obj) {
        if (typeof (obj[key]) === 'object' && obj[key] !== null) {
            clone[key] = preserve ? deepCopy(obj[key], preserve[key]) : deepCopy(obj[key]);
        }
        else {
            clone[key] = preserve && key ? preserve[key] : obj[key];
        }
    }
    return clone;
}
function flattenObject(obj, separator = '_', key = '', map = {}) {
    if (typeof obj === 'object') {
        Object.keys(obj).forEach((k) => {
            const newKey = !key ? k : separator + k;
            flattenObject(obj[k], separator, key += newKey, map);
            key = key.replace(newKey, '');
        });
    }
    else {
        map[key] = true;
    }
    return map;
}
//# sourceMappingURL=useReactiveForm.js.map